/*********************************************************************/
/* neblo.c                                                           */
/* main neblo file: implements the neblo API                         */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#include <neblo_.h>
#include <neblo_macros.h>
#include <neblo_secret.h>
#include <neblo.h>
#include <queues.h>


/* Globals */

u_int16_t           PORT;           /* default peer port               */
unsigned int        MAXCONNS;       /* default max no. of active conns */
unsigned int        MAXSEEDS;       /* no. of seeds locally available  */
struct in_addr      my_ipaddr;      /* IP addr of this host            */
struct in_addr      my_ipaddr_alt;  /* IP addr of "localhost",         */
                                    /* different from the one above.   */
seed_t             *seed;           /* seeds from local seed file      */
ring_t             *ring;           /* peer ring                       */

/* How many fingers a peer can have.  */
unsigned int        NUMFINGERS;

/* Period (secs) of rout tabl maint:  neighbours, and fingers  */
unsigned int        CHK_PERIOD_NEIGHS;
unsigned int        CHK_PERIOD_FINGERS;

/* After this no. of attempts, a neighbour is  considered DEAD         */
unsigned int        NEIGH_DEADLINE;

/* Chk period and deadline (secs) for expected requests */
unsigned int        EXPD_PERIOD;
unsigned int        EXPD_DEADLINE;

/* Chk period and deadline (secs) for pending read msgs */
unsigned int        PEND_READ_PERIOD;
unsigned int        PEND_READ_DEADLINE;

/* Chk period and deadline (secs) for delayed msgs */
unsigned int        RETRY_PERIOD;
unsigned int        RETRY_DEADLINE;

/*
 * Queue of descriptors of incoming connections.
 * A connection persists until one of the partners decides to shut it down,
 * for whatever reason.
 * Once established, an incoming conn carries remote requests to be served
 * by local peer.
 */
reqst_t            *reqst;          /* descriptor queue                   */
reqst_t            *reqst_head;
reqst_t            *reqst_tail;
pthread_mutex_t     reqst_lock;     /* mutex on reqsts list               */
pthread_cond_t      reqst_sem;      /* cond on empty reqsts list          */

int                 wk_s;           /* to wake up sleeping main thread    */
RSA                *pubkey;         /* (ptr to) this peer's public key    */
char               *pubkeymod;      /* holds modulus of pubkey            */
char               *pubkeyexp;      /* holds exponent of pubkey           */
unsigned int        pubkey_size;    /* RSA_size(public key)               */
RSA                *seckey;         /* (ptr to) this peer's secret key    */
FILE               *skf;            /* secret key file, open by connect() */
char               *seckeybuf;      /* buf for reading seckey from file   */
pthread_mutex_t     malloc_lock;    /* mutex on malloc() as well as every */
                                    /* other lib routine using it         */
/*
 * mutex between services that make use of the routing tables, on one hand,
 * and garbage collector of the routing tables, on the other hand.
 */
pthread_mutex_t     clean_lock;
unsigned int        dont_clean;

/* arrays of application-programmed handlers, and related stuff */

pthread_mutex_t     __receiver_handler_lock;
handler            *__receiver_handler_UD;
handler            *__receiver_handler_REQ;
neblo_addr          __refresh_addr;
unsigned int        __refresh_in_progress       = 0;
manager            __refresh_handler            = NULL;
manager            __clean_handler              = NULL;
pthread_mutex_t     __forwarder_handler_lock;
handler            *__forwarder_handler_UD;
handler            *__forwarder_handler_REQ;
handler            *__forwarder_handler_RES;




/*
 *  The law of increasing distances of fingers.
 *  Currently this is a geometrical progression with base 2.  Other
 *  laws (e.g. Fibonacci) are allowed, but a recursive def. is needed.
 *
 *  "magnitude" is the order number of the given finger; nearest one is
 *  the "1st finger" and has magnitude 1 (not 0).  Subsequent fingers
 *  are numbered 2, 3, ... .
 *
 *  Two constraints:  the distance of 1st finger must be <= CUTOFF and
 *  the max allowed distance must not exceed MAXFNGDIST.
 */
neblo_addr
finger_distance_rule
( unsigned int magnitude )
{
  neblo_addr prev;

  if (!magnitude)
    return CUTOFF;
  prev = finger_distance_rule(magnitude-1);
  /*  Watch out for nasty wrap-arounds...  */
  if (prev <= (MAXFNGDIST >> 1))
    return prev + prev;  /*  twice the previous value  */
  /*  Would wrap around or exceed allowed max, so trunc to allowed max  */
  return MAXFNGDIST;
}



/*
 * do_serv: serving cmds and getting data from remote peers
 *          through established connections.
 * 
 * Gets: void *data : ptr to an entry in the circ. list of connection descs.
 *
 * When accepting a remote conn, main thread picks
 * up next elem "E" in the list, fills it with the conn
 * parameters, then unblocks the do_serv() thread
 * associated with "E", which will talk with remote peer.
 * If list is empty, no more conns can be accepted until
 * one active conn shuts down; main threads thus blocks on
 * a condition.
 *
 * When do_serv() detects a conn shutdown, or decides to
 * shut down the conn (after a read() timeout, for instance),
 * the socket is closed and desc "E" is returned back to list;
 * if list was empty (max conns reached), main thread is unblocked.
 * Then, the do_serv() thread related with "E" blocks again.
 *
 * Does not return, unless killed.
 *
 */
void 
*do_serv
( void *data ) 
{
  
  reqst_t               *reqst;
  int                   rc,wkup_main;
  int                   first_time;
  
  
  /* get data passed by the main thread */
  reqst = data;
  wkup_main = 0;

  for (;;) {
    
    pthread_mutex_lock(&reqst->mutx);

    /* stay blocked until main thread will put you at work */
    pthread_cond_wait(&reqst->sem,&reqst->mutx);

    /* unleashed! enter a loop and read the sock until closed */

    pthread_mutex_unlock(&reqst->mutx);
	
#ifdef DEBUG
    S(fprintf(stderr,"[%lu]\tthread unleashed\n",reqst->th_id));
#endif

    for (first_time = 1;;) {

      memset(reqst->buf,0,KEYBUFLEN+SLACK);
      rc = 0;

#ifdef DEBUG
      S(fprintf(stderr,"[%lu]\treading...\n",reqst->th_id));
#endif

      /* wait for a request from peer */
      if (first_time)
        _establishin_(reqst->csock,pubkey,seckey,skf,seckeybuf,&rc);
      else
        _read_(reqst->csock,reqst->buf,KEYBUFLEN,rc,SERV_TIMEOUT,/**/);

#ifdef DEBUG
      S(fprintf(stderr,"[%lu]\t...done (status %d)\n",reqst->th_id,rc));
#endif
	
      /* if connection closed/timed out or something was bad... */
      if (rc < 0) {

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\tconnection closed\n",reqst->th_id));
#endif

        /*
         *  Close sock and return descriptor to desc list.
         *  the _close_ will put csock to NULL.  This will be checked out
         *  by check_routes(), to garbage-collect dead incoming connections
         *  ("friends").
         */

	_close_(&reqst->csock);
	pthread_mutex_lock(&reqst_lock);
        /* if main thread was no longer accepting reqsts... */
	if (reqst_head->next == NULL)
		wkup_main = 1;	       /* ...will unblock it */ 
	reqst->next = NULL;
	reqst_tail->next = reqst;
	reqst_tail = reqst;
	pthread_mutex_unlock(&reqst_lock);
	
	if (wkup_main) {
	  wkup_main = 0;

#ifdef DEBUG
	  S(fprintf(stderr,"[%lu]\twake up main thread\n",reqst->th_id));
#endif

	  pthread_cond_signal(&reqst_sem);
	}

	/* quit read loop, go block again */

#ifdef DEBUG
	S(fprintf(stderr,"[%lu]\tthread blocked again\n",reqst->th_id));
#endif

	break;
      }
      
      if (first_time) {
        /* session key established, go read commands */
        first_time = 0;
        continue;
      }

      /* data or request from a peer */

      /*
       * requests from the communication/routing protocol:
       * write, read, data, refr
       */

      if (!strncmp(reqst->buf,"write ",strlen("write "))) {

#ifdef DEBUG_RW
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_handle_UD_msgs(reqst,FIRST_ATTMPT);
        /* requests of this kind come from "friends", so don't close on fail */

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"read ",strlen("read "))) {

#ifdef DEBUG_RW
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_handle_REQ_msgs(reqst,FIRST_ATTMPT);
        /* requests of this kind come from "friends", so don't close on fail */

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"data ",strlen("data "))) {

#ifdef DEBUG_RW
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_handle_RES_msgs(reqst);
        /*
         *  requests of this kind come from neigh/fingers,
         *  so don't close on fail
         */

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"refr ",strlen("refr "))) {

#ifdef DEBUG_RW
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_handle_refresh_msgs(reqst,FIRST_ATTMPT);
        /* requests of this kind come from "friends", so don't close on fail */

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      /*
       * requests from the connect protocol:
       * register, join, welcome, announce
       */

      if (!strncmp(reqst->buf,"register ",strlen("register "))) {

#ifdef DEBUG_J
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_register(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* close anyway, regardless it went wrong or not */
        _close_(&reqst->csock);
        /* go suspend (will do so, as reqst->csock->sock has been closed) */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"join ",strlen("join "))) {

#ifdef DEBUG_J
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_join(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* don't close, as this cmd travels along routing connections */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"welcome ",strlen("welcome "))) {

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_welcome(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* close anyway, regardless it went wrong or not */
        _close_(&reqst->csock);
        /* go suspend (will do so, as reqst->csock->sock has been closed) */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"announce ",strlen("announce "))) {

#ifdef DEBUG_F
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        rc = neblo_announce(reqst);

        /*
         * requests of this kind travel along maintenance conns,
         * so don't close, even on fail
         */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      /*
       * requests from the routing and route management protocols:
       * friendship, neighlist, findsucc, newsucc, findfinger, newfinger, ping.
       * a "friendship" request is intended to open a permanent connection,
       * whereas "neighlist", "findfinger" and "ping" travel along such
       * permanent connection, so we don't close socks after service
       * (unless something went bad).
       * The requesting peer can however tear down connection on its own side,
       * if needed.
       * Commands "newsucc" and "newfinger", however, travels on temporary
       * connection that are always closed on exit.
       */

      if (!strncmp(reqst->buf,"friendship ",strlen("friendship "))) {

#ifdef DEBUG_F
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_friendship(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* close iff something bad */
        if (rc < 0)
          _close_(&reqst->csock);

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"neighlist",strlen("neighlist"))) {

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_neighlist(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* close iff something bad */
        if (rc < 0)
          _close_(&reqst->csock);

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"findsucc ",strlen("findsucc "))) {

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_findsucc(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* 
         * don't close, as this cmd travels along routing connections.
         * it is up to the neblo_findsucc() routine to decide whether
         * to close or not on some failures.
         */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"newsucc ",strlen("newsucc "))) {

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_newsucc(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* close anyway, regardless it went wrong or not */
        _close_(&reqst->csock);
        /* go suspend (will do so, as reqst->csock->sock has been closed) */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"findfinger ",strlen("findfinger "))) {

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_findfinger(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* 
         * don't close, as this cmd travels along routing connections.
         * it is up to the neblo_findfinger() routine to decide whether
         * to close or not on some failures.
         */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"newfinger ",strlen("newfinger "))) {

#ifdef DEBUG_M
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        pthread_mutex_lock(&clean_lock);
        dont_clean++;
        pthread_mutex_unlock(&clean_lock);

        rc = neblo_newfinger(reqst);

        pthread_mutex_lock(&clean_lock);
        dont_clean--;
        pthread_mutex_unlock(&clean_lock);

        /* close anyway, regardless it went wrong or not */
        _close_(&reqst->csock);
        /* go suspend (will do so, as reqst->csock->sock has been closed) */

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting...\n",reqst->th_id));
#endif

        continue;
      }

      if (!strncmp(reqst->buf,"ping",strlen("ping"))) {

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\t>> %s\n",reqst->th_id,reqst->buf));
#endif

        _write_(reqst->csock,"ping",strlen("ping")+1,rc,/* nothing */);

        /* close iff something bad */
        if (rc < 0)
          _close_(&reqst->csock);

#ifdef DEBUG
        S(fprintf(stderr,"[%lu]\texiting (status %d)...\n",reqst->th_id,rc));
#endif

        continue;
      }

      /* default action = close socket immediately */

#ifdef DEBUG
#if 0
      S(fprintf(stderr,"[%lu] unknown request: %s\n",reqst->th_id,reqst->buf));
#endif
      S(fprintf(stderr,"[%lu] unknown request!!\n",reqst->th_id));
#endif

#if 0
      _close_(&reqst->csock);
      /* go suspend (will do so, as reading from closed sock will fail) */
#endif

    }

  }

  return NULL;  /*  never reached  */

}



/*
 * neblo_listener: listener thread
 * 
 * This thread detects new remote connections and, for each of them, fills
 * up a descriptor (to be taken from a circ queue) and unleash a do_serv()
 * thread.
 *
 * The listener sock has been open and bound to PORT by neblo_init().
 *
 * When max no. of conns is reached, the listener blocks until one of the
 * active conns shut down; the do_serv() instance related to the closing
 * conn will unblock the listener.
 *
 * does not return, unless killed
 *
 */
static void 
*neblo_listener
( void *data )
{

  int			dsize;         /* to select() from listener sock */
  fd_set                rfd;           /* to select() from listener sock */
  fd_set                c_rfd;         /* to select() from listener sock */
  csocket_t            *cs;            /* listener sock */
  csocket_t            *ps;            /* new connection sock */
  struct sockaddr_in	psa;           /* address struct of remote peer */
  unsigned int		size_psa;
  reqst_t              *reqst;         /* desc of incoming connection */
  int                   rc;
  void                 *dummy;


  dummy = data;

  _openin_(INADDR_ANY,&PORT,&cs,&rc);
  /*
   * if PORT was already taken by some other process, the routine above
   * tries with PORT+100, PORT+200, ... until an available port is found.
   * thus, at this point the value of PORT might differ from the one originally
   * established by neblo_init()
   */

  _do_(rc=listen(cs->sock,BACKLOG),rc,"neblo_listener(): listen");

  dsize = getdtablesize();
  size_psa = sizeof(psa);
  /* a set with only one socket open, to receive new incoming connections */
  FD_ZERO(&rfd);
  FD_SET(cs->sock,&rfd);

  /* enter a select-accept-serve infinite loop */
  while (1) {
    
    /*
     * the select() system call waits until any of the file descriptors
     * specified in the read, write and exception sets given to it, is
     * ready to give data, send data, or is in an exceptional state, in
     * respect.  in our case, the latter two ones are NULL.  the call
     * will wait for a given time before returning. in this case, the
     * value is NULL, so it will not timeout. dsize specifies the size of
     * the file descriptor table. on return, each set of file
     * descriptors is updated to contain only those descriptors whose
     * state has changed.
     */
    
    c_rfd = rfd;
    _do_(rc = select(dsize, &c_rfd, NULL, NULL, (struct timeval *)NULL),
	 rc<0, "neblo_listener(): select");
    if (FD_ISSET(cs->sock, &c_rfd)) {
      /* new connection requested */
      
      /* SANITY chk */
      if (reqst_head->next == NULL) {
	perror("unknown err"); continue;
      }
      
      _accept_(cs,(struct sockaddr *)&psa,&size_psa,&ps,&rc);
      if (rc < 0)
        continue;
      
      /* begin serving this request: detach descriptor from head */
      reqst = reqst_head;
      reqst_head = reqst_head->next;
      /* fill in the request descriptor */
      reqst->csock = ps;
      reqst->peer_addr.s_addr = psa.sin_addr.s_addr;

#ifdef DEBUG
      S(fprintf(stderr,"new connection from %u.%u.%u.%u --> [%lu]\n",
              ((unsigned char *)&reqst->peer_addr)[0],
              ((unsigned char *)&reqst->peer_addr)[1],
              ((unsigned char *)&reqst->peer_addr)[2],
              ((unsigned char *)&reqst->peer_addr)[3],
              reqst->th_id));
#endif
      
      /* unleash the thread associated to descriptor */
      _do_(rc = pthread_cond_signal(&reqst->sem),rc,
	   "neblo_listener(): cond_signal");
      /* new connection is now accepted */
      if (reqst_head->next == NULL) {
	/* desc queue is empty, max no. of pending reqsts reached. */
	/* suspend thread; resume as soon as a reqst completes */

#ifdef DEBUG
	S(fprintf(stderr,"max connections reached. suspend...\n"));
#endif

        pthread_mutex_lock(&reqst_lock);
        pthread_cond_wait(&reqst_sem,&reqst_lock);
        pthread_mutex_unlock(&reqst_lock);

#ifdef DEBUG
	S(fprintf(stderr,"resume accepting connections\n"));
#endif

      }
      continue;
    }
    
  }

}



/*
 * neblo_sigpipehandler : signal handler for catching and ignoring SIGPIPE.
 * if this were not done, the peer would die by SIGPIPE whenever a remote
 * partner closes an active connection.
 * 
 * Gets:
 *   int signum: signal no. that caused the handler to run
 *
 * Returns: void
 *
 */
static void
neblo_sigpipehandler
( int signum )
{

#ifdef DEBUG
  fprintf(stderr,"catched a SIGPIPE\n");
#endif

  signal(signum,&neblo_sigpipehandler);
  return;
}



/*
 * neblo_submit:  the invoker submits a connect req. for the ring of peers.
 * Returns 0 if success, < 0 otherwise.
 */
int 
neblo_submit
( void )
{
  
  int                   rc;
  csocket_t            *cs;
  unsigned int          j;
  char                 *cmd;
  neblo_addr            n_A;
  neblo_addr            n_B;
  unsigned char        *n_A_k,*n_B_k;
  char                 *keymod;
  char                 *keyexp;
  peer_t               *f;
  char                 *myhostname;
  struct hostent       *haddr;


  if (!MAXSEEDS) {
    /* this peer runs in isolation. will create ring by itself */

    /* 64 bit id. of this peer */
    RAND_bytes((unsigned char *)&ring->my_id_src,sizeof(ring->my_id));
    _do_(_malloc_(n_A_k,RIPEMD160_DIGEST_LENGTH+SLACK),
         !n_A_k, "neblo_submit(): n_A_k malloc (II)");
    RIPEMD160((unsigned char *)&ring->my_id_src,sizeof(ring->my_id_src),n_A_k);
    /*  Fold 160 bits over 64 bits  */
    xor_fold((int8_t *)&ring->my_id,sizeof(ring->my_id),
             (int8_t *)n_A_k,RIPEMD160_DIGEST_LENGTH);
    _free_(n_A_k);

    /*  All addresses are initially allocated to this peer.  */
    /*  Lower bound of addr interval is my_id itself.  */
    ring->up_bound = ring->my_id - 1;

    /*
     *  Finger list remains empty.  Neighbour list, however, must contain a
     *  copy of myself, so that the ring is closed; otherwise, routing would
     *  not be able to wrap along the ring.  Predecessor is not allocated
     *  (and, if it was, it get disposed now); it will once next peer is
     *  welcomed.
     */
    _do_(_calloc_(f,sizeof(peer_t)),!f,"neblo_submit(): f calloc");
    _pthmutexinit_(&f->lock);
    /*  Compute the IP addr of local host...  */
    _do_(_malloc_(myhostname,BUFLEN+SLACK),!myhostname,
         "neblo_submit(): myhostname malloc");
    gethostname(myhostname,BUFLEN);
    haddr = gethostbyname(myhostname);
    _free_(myhostname);
    if (haddr == NULL) {
      fprintf(stderr,
              "neblo_submit(): gethostbyname() returned NULL.  abort...\n");
      exit(1);
    }
    memcpy(&f->addr.s_addr,haddr->h_addr_list[0],haddr->h_length);
    f->port = PORT;
    f->pubkey = pubkey;
    f->status = TOCALL;
    /*  Take note that this entry is myself actually.  */
    f->self = 1;
    f->next = NULL;
    ring->neigh_list = f;

#ifdef VERBOSE
    fprintf(stderr,
            "neblo_submit(): isolation: id %016llx, key intvl %016llx - %016llx\n",
            ring->my_id,ring->my_id,ring->up_bound);
#endif

    ring->enabled = 1;

    return 0;
  }

  /* not an isolated peer. that is, it will send a "register" req. to a seed */

  _do_(_malloc_(cmd,KEYBUFLEN+SLACK), !cmd, "neblo_submit(): cmd malloc");
  _do_(_malloc_(n_A_k,RIPEMD160_DIGEST_LENGTH+SLACK),
       !n_A_k, "neblo_submit(): n_A_k malloc");
  _do_(_malloc_(n_B_k,RIPEMD160_DIGEST_LENGTH+SLACK),
       !n_B_k, "neblo_submit(): n_B_k malloc");

  /* choose a seed peer, send it a "register" req for required ring */
  for (j=0;j<MAXSEEDS;j++) {
      
#ifdef DEBUG
    fprintf(stderr,
            "neblo_submit(): sending register req to peer %u.%u.%u.%u:%u\n",
            ((unsigned char *)&seed[j].addr.s_addr)[0],
            ((unsigned char *)&seed[j].addr.s_addr)[1],
            ((unsigned char *)&seed[j].addr.s_addr)[2],
            ((unsigned char *)&seed[j].addr.s_addr)[3],
            seed[j].port);
#endif

    _openout_(seed[j].addr,(u_int16_t)seed[j].port,&cs,seed[j].pubkey,&rc);
    if (rc < 0)
      /*  failed to contact that seed.  try another one  */
      continue;

    /*  Set up a handle for the "register" cmd (which is recursive-style)  */
    cmd_handle_set(&ring->join_handle);

    /*
     *  Issue a "register" request, so starting the connect protocol.
     *  See file register.c for a description.
     */

    _do_(_BN_bn2hex_(keymod,pubkey->n),!keymod,"neblo_submit(): keymod alloc");
    _do_(_BN_bn2hex_(keyexp,pubkey->e),!keyexp,"neblo_submit(): keyexp alloc");
    sprintf(cmd,"register %u %u %s %s",
            PORT,cmd_handle_val(&ring->join_handle),keymod,keyexp);
    _OPENSSL_free_(keymod); _OPENSSL_free_(keyexp);
    _write_(cs,cmd,strlen(cmd)+1,rc,continue);

#ifdef DEBUG
    fprintf(stderr, "\t<< %s\n",cmd);
#endif

    /* generate 64 bit random number "A" */
    RAND_bytes((unsigned char *)&n_A,sizeof(n_A));

    /* send RIPEMD160(A) to seed */
    RIPEMD160((unsigned char *)&n_A,sizeof(n_A),n_A_k);
    _write_(cs,n_A_k,RIPEMD160_DIGEST_LENGTH,rc,continue);

#ifdef DEBUG
    fprintf(stderr, "\t<< (%u) %u %u %u %u %u\n",
            RIPEMD160_DIGEST_LENGTH,
            (unsigned int)n_A_k[0], (unsigned int)n_A_k[4],
            (unsigned int)n_A_k[8], (unsigned int)n_A_k[12],
            (unsigned int)n_A_k[16]);
#endif

    /* get RIPEMD160(B) from seed */
    _read_(cs,n_B_k,RIPEMD160_DIGEST_LENGTH,rc,TIMEOUT,continue);

#ifdef DEBUG
    fprintf(stderr, "\t>> (%u) %u %u %u %u %u\n",
            RIPEMD160_DIGEST_LENGTH,
            (unsigned int)n_B_k[0], (unsigned int)n_B_k[4],
            (unsigned int)n_B_k[8], (unsigned int)n_B_k[12],
            (unsigned int)n_B_k[16]);
#endif

    /* send A */
    _write_(cs,&n_A,sizeof(n_A),rc,continue);

#ifdef DEBUG
    fprintf(stderr, "\t<< A was %016llx\n",n_A);
#endif

    /* get B */
    _read_(cs,&n_B,sizeof(n_B),rc,TIMEOUT,continue);

#ifdef DEBUG
    fprintf(stderr, "\t>> B was %016llx\n",n_B);
#endif

    /* check RIPEMD160(B) against the original "B". abort on fail */
    RIPEMD160((unsigned char *)&n_B,sizeof(n_B),n_A_k);
    if (strncmp((char *)n_B_k,(char *)n_A_k,RIPEMD160_DIGEST_LENGTH)) {

#ifdef DEBUG
      fprintf(stderr,"\tcheck failed!\n");
#endif

      _close_(&cs);
      continue;
    }

    /* compute the unique id. of this peer */
    ring->my_id_src = n_A ^ n_B;
    RIPEMD160((unsigned char *)&ring->my_id_src,sizeof(ring->my_id_src),n_A_k);
    /*  Fold 160 bits over 64 bits  */
    xor_fold((int8_t *)&ring->my_id,sizeof(ring->my_id),
             (int8_t *)n_A_k,RIPEMD160_DIGEST_LENGTH);

#ifdef VERBOSE
    fprintf(stderr,"\tmy id is %016llx, source %016llx\n",
            ring->my_id,ring->my_id_src);
#endif

    /*
     *  At this point, the seed is starting phase 2 of connect protocol.
     *  A "welcome" command might eventually arrive; it will be served
     *  by another thread, function neblo_welcome(), which will reset
     *  the "ring->join_handle" when finished, so we poll "ring->join_handle"
     *  to see when this happens (or to decide a timeout).
     */

    rc = 0;

    for (;;) {
      sleep(1);
      cmd_handle_lock(&ring->join_handle);
      if (cmd_handle_is_done(&ring->join_handle))
        /*  "welcome" has arrived  */
        break;
      if (cmd_handle_age(&ring->join_handle) > (int)REC_CMD_DEADLINE) {
        /*  Timed out.  Abort; caller will try another seed.  */
        cmd_handle_reset(&ring->join_handle);
        rc = -1;
        break;
      }
      cmd_handle_unlock(&ring->join_handle);
    }
    cmd_handle_unlock(&ring->join_handle);

    if (rc >= 0)
      break;

  }
    
  _free_(cmd);
  _free_(n_A_k); _free_(n_B_k);

  return rc;
  
}



/*
 * neblo_init : peer bootup.
 * mask SIGPIPE off, read cmd line, read cfg file, read seed file,
 * alloc/init globals
 * 
 * Gets: int argc, char** argv :
 *   -f <fname> : file "fname" stores config file
 *   -s <fname> : file "fname" stores seeds file
 *   -p <fname> : file "fname" stores this peer's public key
 *   -k <fname> : file "fname" stores this peer's private key
 *
 */
void 
neblo_init
( int argc, char **argv )
{

  FILE                 *cfg,*sd,*pk;
  int                   rc;
  unsigned int          i;
  char                 *myhostname;
  struct hostent       *haddr;	        /* host-to-IP translation */
  char                 *line,*buf1;
  char                 *buf2,*buf3;
  char                 *pkmod;
  char                 *pkexp;
  char                 *nam1,*nam2;
  char                 *nam3,*nam4;
  int                   found_f;
  int                   found_s;
  int                   found_p;
  int                   found_k;
  RSA                  *seedkey;       /* to read pub keys of seeds */
  struct timeval        tv;
  double                val;
  
  
  signal(SIGPIPE,&neblo_sigpipehandler);

  _do_(line = (char *)malloc(LINESIZE+SLACK),!line,"init(): line malloc");
  _do_(buf1 = (char *)malloc(BUFLEN+SLACK),!buf1,"init(): buf1 malloc");
  _do_(buf2 = (char *)malloc(BUFLEN+SLACK),!buf2,"init(): buf2 malloc");
  _do_(buf3 = (char *)malloc(BUFLEN+SLACK),!buf3,"init(): buf3 malloc");
  _do_(pkmod = (char *)malloc(KEYBUFLEN+SLACK),!pkmod,"init(): pkmod malloc");
  _do_(pkexp = (char *)malloc(BUFLEN+SLACK),!pkexp,"init(): pkexp malloc");
  _do_(nam1 = (char *)malloc(BUFLEN+SLACK),!nam1,"init(): nam1 malloc");
  _do_(nam2 = (char *)malloc(BUFLEN+SLACK),!nam2,"init(): nam2 malloc");
  _do_(nam3 = (char *)malloc(BUFLEN+SLACK),!nam3,"init(): nam3 malloc");
  _do_(nam4 = (char *)malloc(BUFLEN+SLACK),!nam4,"init(): nam4 malloc");
  _do_(seedkey = RSA_new(),!seedkey,"init(): RSA_new of seedkey");
  
  fprintf(stderr,"================================================================================\n");

  /* alloc this peer's pub/secr key pair */
  _do_(pubkey = RSA_new(),!pubkey,"RSA_new() failed (pubkey)");
  _do_(pubkeymod = (char *)malloc(KEYBUFLEN+SLACK),!pubkeymod,
       "init(): pubkeymod malloc");
  _do_(pubkeyexp = (char *)malloc(BUFLEN+SLACK),!pubkeyexp,
       "init(): pubkeyexp malloc");
  /*
   * NOTE: secret key is not allocated initially.  it is only allocated when
   * needed (macro _get_private_key_), and immediately erased and disposed
   * afterwards (macro _clean_private_key_).  Here we only allocate the
   * buffer for reading the secret key from file.
   */
  _do_(seckeybuf = (char *)malloc(KEYBUFLEN+SLACK),!seckeybuf,
       "init(): seckeybuf malloc");
  
  /*  Defaults for commons.  */
  PORT = 8000;
  MAXCONNS = 10;
  MAXSEEDS = 100;

  /*  Get IP addr of this host and of "localhost".  */
  _do_(myhostname = malloc(BUFLEN+SLACK),!myhostname,
       "init(): myhostname malloc");
  gethostname(myhostname,BUFLEN);
  _do_(haddr = gethostbyname(myhostname),!haddr,
       "init(): could not determine local host IP addr");
  free(myhostname);
  memcpy(&my_ipaddr.s_addr,haddr->h_addr_list[0],haddr->h_length);
  _do_(haddr = gethostbyname("localhost"),!haddr,
       "init(): could not determine IP addr of localhost");
  memcpy(&my_ipaddr_alt.s_addr,haddr->h_addr_list[0],haddr->h_length);

  CHK_PERIOD_NEIGHS = 10;      /* secs */
  CHK_PERIOD_FINGERS = 20;     /* secs */
  NEIGH_DEADLINE = 2;          /* attempts; approx.                       */
                               /* NEIGH_DEADLINE*TIMEOUT secs             */
  EXPD_PERIOD = 60;            /* secs */
  EXPD_DEADLINE = 120;         /* secs; must be far greater than          */
                               /* CHK_PERIOD_NEIGHS, otherwise new peers  */
                               /* will not be able to establish contacts  */
                               /* with their neighbours.                  */
  PEND_READ_PERIOD = 15;       /* secs */
  PEND_READ_DEADLINE = 15;     /* secs */
  RETRY_PERIOD   = 10;         /* secs */
  RETRY_DEADLINE = 60;         /* secs */

  /* init the random generator */
  gettimeofday(&tv,NULL);
  srand(tv.tv_usec);
  do {
    val = rand();
    RAND_seed((void *)&val,sizeof(val));
  } while (!RAND_status());

  /* scan cmd line for options -f -s -p -k */
  found_f = found_s = found_p = found_k = 0;
  for (;;) {
    while ((rc = getopt(argc,argv,"f:s:p:k:")) != 'f' && rc != 's' && rc != 'p' && rc != 'k' && rc != -1);
    if (rc == -1 || rc == '?')
      break;
    if (rc == 'f' && !found_f) {
      found_f = 1;
      strncpy(nam1,optarg,BUFLEN);
      strcat(nam1,"\0"); /* just in case optarg exceeds BUFLEN chars */
    }
    if (rc == 's' && !found_s) {
      found_s = 1;
      strncpy(nam2,optarg,BUFLEN);
      strcat(nam2,"\0"); /* just in case optarg exceeds BUFLEN chars */
    }
    if (rc == 'p' && !found_p) {
      found_p = 1;
      strncpy(nam3,optarg,BUFLEN);
      strcat(nam3,"\0"); /* just in case optarg exceeds BUFLEN chars */
    }
    if (rc == 'k' && !found_k) {
      found_k = 1;
      strncpy(nam4,optarg,BUFLEN);
      strcat(nam4,"\0"); /* just in case optarg exceeds BUFLEN chars */
    }
  }
  
  /*
   * get values for PORT and MAXCONNS for this peer from
   * file "nam1".  use default config file if "nam1" not found.
   * use default values if default config file not found.
   */
  if (!found_f) {

#ifdef VERBOSE
    fprintf(stderr,"default config file: %s\n",DEFAULT_CFG_FILE);
#endif

    cfg = fopen(DEFAULT_CFG_FILE,"r");
  } else {

#ifdef VERBOSE
    fprintf(stderr,"config file: %s\n",nam1);
#endif

    cfg = fopen(nam1,"r");
    if (cfg == NULL) {
      fprintf(stderr,"could not open config file %s. will use %s instead\n",
              nam1,DEFAULT_CFG_FILE);
      cfg = fopen(DEFAULT_CFG_FILE,"r");
    }
  }
  if (cfg == NULL) {
    fprintf(stderr,"could not open default config file %s. using defaults.\n",
            DEFAULT_CFG_FILE);
  } else {
    while ((rc = (int)fgets(line,LINESIZE,cfg))) {
      /* format of cfg file is: list of <id val> pairs. */
      /* empty the strings before parsing line */
      buf1[0] = buf2[0] = 0;
      sscanf(line,"%100s %100s",buf1,buf2);

      /* chk if line in config file is a comment, possibly discard */
      if (buf1[0] == '#') {
        _discardline_(line,cfg);
	continue;
      }

      rc = atoi(buf2);
      sprintf(buf3,"%d",rc);

      if (!strcmp(buf1,"PORT")) {
        /* chk if 2nd field is a valid port no., possibly discard */
        if (strcmp(buf2,buf3) || rc < 8000) {
          _discardline_(line,cfg);
	  continue;
        }
        PORT = rc;
        continue;
      }

      if (!strcmp(buf1,"MAXCONNS")) {
	/* chk if 2nd field is a positive no., possibly discard */
	if (strcmp(buf2,buf3) || rc <= 0) {
          _discardline_(line,cfg);
	  continue;
	}
	MAXCONNS = rc;
	continue;
      }

      /* go to next line */
    }
    fclose(cfg);
  }

#ifdef VERBOSE
  fprintf(stderr,"PORT %d, MAXCONNS %d\n",PORT,MAXCONNS);
#endif
  
  /*
   * alloc and fill array of seeds from file "nam2".
   * will use default seed file if "nam2" not found.
   * will run in isolation mode if no seeds in file or default file not found.
   */
  _do_(seed = (seed_t *)malloc(MAXSEEDS*sizeof(seed_t)),!seed,
       "init(): seed malloc");
  memset(seed,0,MAXSEEDS * sizeof(seed_t));
  if (!found_s) {

#ifdef VERBOSE
    fprintf(stderr,"default seed file: %s\n",DEFAULT_SEEDS_FILE);
#endif

    sd = fopen(DEFAULT_SEEDS_FILE,"r");
  } else {

#ifdef VERBOSE
    fprintf(stderr,"seed file: %s\n",nam2);
#endif

    sd = fopen(nam2,"r");
    if (sd == NULL) {
      fprintf(stderr,"could not open seed file %s. will use %s instead\n",
              nam2,DEFAULT_SEEDS_FILE);
      sd = fopen(DEFAULT_SEEDS_FILE,"r");
    }
  }
  if (sd == NULL) {
    /* no seed file.  empty seed array */
    fprintf(stderr,"could not open default seed file %s."
            " starting in isolation\n",DEFAULT_SEEDS_FILE);
    MAXSEEDS = 0;
  } else {
    i = 0;
    while ((rc = (int)fgets(line,LINESIZE,sd))) {
      /*
       * format of seed file is: list of <hostname port pubkeymod pubkeyexp>
       * triples
       */
      /* empty the strings before parsing line */
      buf1[0] = buf2[0] = pkmod[0] = pkexp[0] = 0;
      sscanf(line,"%100s %100s %50000s %100s",buf1,buf2,pkmod,pkexp);

#ifdef DEBUG
      fprintf(stderr,"%s ||| %s ||| %s ||| %s\n",buf1,buf2,pkmod,pkexp);
#endif

      /* chk if line in config file is a comment, possibly discard */
      if (buf1[0] == '#') {
        _discardline_(line,sd);
        continue;
      }

      /* chk if first field is a valid host, possibly discard */
      if (!(haddr = gethostbyname(buf1))) {
        _discardline_(line,sd);
        continue;
      }

      rc = atoi(buf2);
      sprintf(buf3,"%d",rc);

      /* chk if 2nd field is a valid port no., possibly discard */
      if (strcmp(buf2,buf3) || rc < 8000) {
        _discardline_(line,sd);
        continue;
      }

      _convert_pubkey_(seedkey,pkmod,pkexp,/* nothing */);
       /* chk if <3rd 4th> fields were a legal key, possibly discard */
      if (!(seedkey->n && seedkey->e)) {
        _discardline_(line,sd);
        continue;
      } 

      /* this is a valid seed. save it in the array */
      memcpy(&seed[i].addr.s_addr,haddr->h_addr_list[0],haddr->h_length);
      seed[i].port = rc;
      seed[i].pubkey = seedkey;
      _do_(seedkey = RSA_new(),!seedkey,"init(): RSA_new of seedkey");
    
      i++;
      if (i >= MAXSEEDS)
        break;

      /* go to next line */
    }
    fclose(sd);
    MAXSEEDS = i;
  }
  
#ifdef VERBOSE
  if (!MAXSEEDS)
    fprintf(stderr,"no seeds in file. starting in isolation\n");
  else
    for (i=0;i<MAXSEEDS;i++)
      fprintf(stderr,"%u.%u.%u.%u:%u\n",
              ((unsigned char *)&seed[i].addr.s_addr)[0],
              ((unsigned char *)&seed[i].addr.s_addr)[1],
              ((unsigned char *)&seed[i].addr.s_addr)[2],
              ((unsigned char *)&seed[i].addr.s_addr)[3],
              seed[i].port);
#endif

  /*
   * get public key from file "nam3", and store it into global *pubkey
   * will use default pubkey file if "nam3" not found.
   * will generate a default public/private key pair if default pubkey file
   * not found.
   */
  if (!found_p) {

#ifdef VERBOSE
    fprintf(stderr,"default pubkey file: %s\n",DEFAULTPUBKEYFILE);
#endif

    pk = fopen(DEFAULTPUBKEYFILE,"r");
  } else {

#ifdef VERBOSE
    fprintf(stderr,"pubkey file: %s\n",nam3);
#endif

    pk = fopen(nam3,"r");
    if (pk == NULL) {
      fprintf(stderr,"could not open pubkey file %s. will use %s instead\n",
              nam3,DEFAULTPUBKEYFILE);
      pk = fopen(DEFAULTPUBKEYFILE,"r");
    }
  }
  if (pk == NULL) {
    /*
     * no pubkey file.  try to generate a pub/priv key pair and store them
     * under default file names
     */
    fprintf(stderr,"could not open default pubkey file %1$s.\n"
            "\twill attempt to generate a default public/secret key pair,\n"
            "\tstoring them into files %1$s and %2$s.\n"
            "\tnow running utility \"makeys\"...\n",
            DEFAULTPUBKEYFILE, DEFAULTSECKEYFILE);
    if (system("./makeys > /dev/null")) {
      fprintf(stderr,"\t\tfailed to run utility \"makeys\". bailing out\n");
      goto abort;
    }
    fprintf(stderr,"\tmakeys done.\n");
    /* key pair now stored under default file names.  open pubkey file */
    pk = fopen(DEFAULTPUBKEYFILE,"r");
    if (pk == NULL) {
      fprintf(stderr,"\t\tfailed to open default pubkey file. bailing out\n");
      goto abort;
    }
  }
  /* format of pubkey file is: pubkeymod \n pubkeyexp */
  /* empty the strings before reading file */
  pubkeymod[0] = pubkeyexp[0] = 0;
  while ((rc = fscanf(pk,"%50000s\n%100s",pubkeymod,pubkeyexp)) >= 0) {
    /* chk if line in config file is a comment, possibly discard */
    if (pubkeymod[0] == '#')
      continue;
    _convert_pubkey_(pubkey,pubkeymod,pubkeyexp,continue);
    /* public key was OK, so quit reading file */
    break;
  }
  fclose(pk);
  /* sanity: chk if we finally got a legal public key, possibly abort */
  if (!(pubkey->n && pubkey->e)) {
    fprintf(stderr,"could not get a public key for this peer. bailing out\n");
    goto abort;
  }
  pubkey_size = RSA_size(pubkey);

  /*
   * test (but will not keep in memory, for privacy reasons) private key from
   * file "nam4".
   * will use default seckey file if "nam4" not found.
   * will abort if default seckey file not found.
   * will leave seckey file open for future accesses (global FILE *skf).
   */
  if (!found_k) {

#ifdef VERBOSE
    fprintf(stderr,"default secret key file: %s\n",DEFAULTSECKEYFILE);
#endif

    skf = fopen(DEFAULTSECKEYFILE,"r");
  } else {

#ifdef VERBOSE
    fprintf(stderr,"secret key file: %s\n",nam4);
#endif

    skf = fopen(nam4,"r");
    if (skf == NULL) {
      fprintf(stderr,"could not open secret key file %s. will use %s instead\n",
              nam4,DEFAULTSECKEYFILE);
      skf = fopen(DEFAULTSECKEYFILE,"r");
    }
  }
  if (skf == NULL) {
    /* no seckey file.  it should rather have been found!  abort... */
    fprintf(stderr,"could not open default secret key file %s. bailing out\n",
            DEFAULTSECKEYFILE);
    goto abort;
  }

  /* chk if this peer's public/secret key pair works */
  {
    char *test_msg = "morning glory";
    char *encr_msg, *decr_msg;
    int len;

    len = strlen(test_msg)+1;
    _do_(encr_msg = (char *)malloc(pubkey_size),!encr_msg,
         "init(): encr_msg malloc");
    _do_(decr_msg = (char *)calloc(1,len),!decr_msg,
         "init(): decr_msg calloc");
    rc = RSA_public_encrypt(len, (unsigned char *)test_msg,
                            (unsigned char *)encr_msg, pubkey,
                            RSA_PKCS1_OAEP_PADDING);
    _get_private_key_(skf,seckey,seckeybuf);
    rc *= RSA_private_decrypt(pubkey_size, (unsigned char *)encr_msg,
                             (unsigned char *)decr_msg, seckey,
                             RSA_PKCS1_OAEP_PADDING);
    _clean_private_key_(seckey,seckeybuf,KEYBUFLEN+SLACK);
    if (strcmp(decr_msg,test_msg))
      rc *= -1;
    free(encr_msg);
    free(decr_msg);
    if (rc < 0) {
      fprintf(stderr, "init(): public/secret key pair is broken. aborting\n");
      goto abort;
    }
  }

  /*
   *  seckey file has been found, has been open and will be left so for
   *  future accesses.  global FILE *skf is entry point
   */

  /* init pthread locks and condition variables */
  _pthmutexinit_(&reqst_lock);
  _pthcondinit_(&reqst_sem);
  _pthmutexinit_(&malloc_lock);

  _pthmutexinit_(&clean_lock);
  dont_clean = 0;

  /*
   *  alloc circ. list of MAXCONNS+1 descriptors for incoming connections.
   *  "+1" to avoid head/tail overrun.
   *  spawn MAXCONNS+1 serving threads, each will block on private sem
   */
  _do_(reqst = (reqst_t *)malloc((MAXCONNS+1)*sizeof(reqst_t)),!reqst,
       "main(): malloc");

  for (i=0;i<=MAXCONNS;i++) {
    reqst[i].csock = NULL;
    _do_(reqst[i].buf = (char *)malloc(KEYBUFLEN+SLACK),!reqst[i].buf,
         "main(): malloc");
    _pthmutexinit_(&reqst[i].mutx);
    _pthcondinit_(&reqst[i].sem);

    _do_(rc = pthread_create(&reqst[i].th_id,NULL,do_serv,&reqst[i]),rc,
         "main(): pthread_create");
    /* thread will immediately block on reqst[i].sem */

    reqst[i].next = &reqst[i+1];
  }

  reqst[MAXCONNS].next = NULL;
  reqst_head = &reqst[0];
  reqst_tail = &reqst[MAXCONNS];

  /* init the list of expected requests */
  init_expd_reqst_q();

  /* init the list of pending read messages */
  init_pend_read_q();

  /* init the ring of peers */

  _do_(ring = (ring_t *)calloc(1,sizeof(ring_t)),!ring,"init(): ring calloc");
  _pthmutexinit_(&ring->lock);
  _pthmutexinit_(&ring->write_lock);
  _pthmutexinit_(&ring->join_lock);
  cmd_handle_init(&ring->join_handle);
  /*  Everything else, incl. "ring->enabled", is 0/NULL by "calloc".  */

  /*
   *  Calculating how many fingers we can have, based on the recursive
   *  rule finger_distance_rule() (above in this file) defining the placement
   *  of fingers.
   */
  for (NUMFINGERS=1; finger_distance_rule(NUMFINGERS) < MAXFNGDIST;
       NUMFINGERS++);
  if (finger_distance_rule(NUMFINGERS) > MAXFNGDIST)
    NUMFINGERS--;

  /*
   *  Alloc and init finger table.  This is an array.  Neighbours and friends,
   *  however, are lists (num. of elements not known in advance), allocated on
   *  demand.
   *  "calloc" will fill every peer desc with 0/NULL, so the initial state of
   *  each finger is NONE.
   */
  _do_(ring->finger_table = (peer_t *)calloc(NUMFINGERS,sizeof(peer_t)),
       !ring->finger_table,"init(): finger_table calloc");
  for (i=0; i<NUMFINGERS; i++) {
    _pthmutexinit_(&ring->finger_table[i].lock);
    cmd_handle_init(&ring->finger_table[i].handle);
  }
      
  /*
   *  Table of msg handlers.
   *  Each table is an array of 65536 ptrs to functions.
   */

  _do_(__receiver_handler_UD = (handler *)calloc(65536,sizeof(handler)),
       !__receiver_handler_UD,"init(): __receiver_handler_UD calloc");
  _do_(__receiver_handler_REQ = (handler *)calloc(65536,sizeof(handler)),
       !__receiver_handler_REQ,"init(): __receiver_handler_REQ calloc");

  _do_(__forwarder_handler_UD = (handler *)calloc(65536,sizeof(handler)),
       !__forwarder_handler_UD,"init(): __forwarder_handler_UD calloc");
  _do_(__forwarder_handler_REQ = (handler *)calloc(65536,sizeof(handler)),
       !__forwarder_handler_REQ,"init(): __forwarder_handler_REQ calloc");
  _do_(__forwarder_handler_RES = (handler *)calloc(65536,sizeof(handler)),
       !__forwarder_handler_RES,"init(): __forwarder_handler_RES calloc");

  _pthmutexinit_(&__receiver_handler_lock);
  _pthmutexinit_(&__forwarder_handler_lock);
  

  /* end of initialization */

  _free_(line); _free_(buf1); _free_(buf2); _free_(buf3);
  _free_(pkmod); _free_(pkexp);
  _free_(nam1); _free_(nam2); _free_(nam3); _free_(nam4);
  _RSA_free_(seedkey);

  /* be "sure" all threads block on their sems...  */
  sleep(1);

  return;

abort:
  _free_(line); _free_(buf1); _free_(buf2); _free_(buf3);
  _free_(pkmod); _free_(pkexp);
  _free_(nam1); _free_(nam2); _free_(nam3); _free_(nam4);
  _RSA_free_(seedkey);
  exit(1);

}



/*
 * neblo_connect(): register the peer with the overlay; start listener thread,
 *                  server threads, and routing table maintenance thread
 */
int 
neblo_connect
( void )
{
  
  int			rc; 		/* system calls return value storage */
  pthread_t             th_id;
  unsigned int          count;
  

  /* start the listener thread */
  _do_(rc = pthread_create(&th_id,NULL,neblo_listener,NULL),rc,
       "neblo_connect(): neblo_listener() pthread_create");

  /* be "sure" listener has started, before starting a connect phase */
  sleep(1);

  /* Connect.  Max 10 attempts. */
  for (count=10; count; count--) {
    if (neblo_submit() >= 0)
      break;
    sleep(1);
  }

  if (!count) {
    fprintf(stderr, "neblo_connect(): could not connect.  abort...\n");
    exit(0);
  }

  /*
   *  Initialize the "smoke".  A random value not larger than CUTOFF/4.
   *  To know how to use the "smoke", see findfinger.c
   */
  pthread_mutex_lock(&ring->lock);
  RAND_bytes((unsigned char *)&ring->smoke,sizeof(ring->smoke));
  for (; ring->smoke > CUTOFF<<2; ring->smoke >>= 1);
  pthread_mutex_unlock(&ring->lock);

  /*  Start the garb coll for the queue of expected requests.  */
  _do_(rc = pthread_create(&th_id,NULL,clean_expd_reqsts,(void *)NULL),rc,
       "neblo_connect(): clean_expd_reqsts() pthread_create");

  /*  Start the garb coll for the queue of pending read requests.  */
  _do_(rc = pthread_create(&th_id,NULL,clean_pend_reads,NULL),rc,
       "neblo_connect(): clean_pend_reads() pthread_create");

  /*  Start the garb coll for the queue of delayed msgs.  */
  _do_(rc = pthread_create(&th_id,NULL,clean_retry_queue,NULL),rc,
       "neblo_connect(): clean_retry_queue() pthread_create");

  /*  Start the routing tables maintenance threads.  */

  _do_(rc = pthread_create(&th_id,NULL,neblo_check_neighbours,&th_id),rc,
       "neblo_connect(): neblo_check_neighbours() pthread_create");

  _do_(rc = pthread_create(&th_id,NULL,neblo_check_fingers,&th_id),rc,
       "neblo_connect(): neblo_check_fingers() pthread_create");

  /*  Trigger the first creation of finger table.  */
  finger_create(th_id);  /* same th_id as the check_finger() */

  return 0;

}



/*
 * Send "cp" copies of a write msg resp. to the peers who map addresses
 * "addr", "addr+2^(N-1)", "addr+2^(N-1)+2^(N-2)", ...,
 * "addr+2^(N-1)+2^(N-2)+...+2^(N-cp+1)"
 * where N is the addr size in bit (64 bit in our case).
 *
 * The msg copies are sent by successive recursive calls.  First invocation
 * must have "depth" = 1.  Each recursive call increases "depth" by 1.
 *
 * Returns -k where k is the no. of failed copies (k <= cp), so 0 on success.
 * "Failed" means it has been impossible to route msg outside of local peer;
 * they will however retried later automatically.
 *
 * Msg size cannot exceed NEBLO_MSGSIZE.
 *
 * Since it is a UniDirectional msg, do not expect any answer.
 *
 * Semantics is "blocking".
 *
 * NOTE:  this is a support routine.  Application cannot invoke this
 * directly; they should rather use neblo_UD_send() or neblo_UD_send_h()
 * (see below)
 */
static int
_neblo_UD_send
(
  unsigned int cp, neblo_port port, neblo_addr addr,
  neblo_addr real_addr, void *msg, unsigned int len, unsigned int depth,
  int from_refresh_handler
)
{

  unsigned char        *buf;
  peer_t               *hop;
  int                   rc = 0;
  reqst_t              *req;
  unsigned int          tmpunavail = 0;


  _do_(_calloc_(buf,HEADERSIZE_W+len+SLACK),!buf,
       "_neblo_UD_send(): buf calloc");
  len = min(len,NEBLO_MSGSIZE);

  for (;;) {

    hop = next_hop(addr);

    if (hop == NULL) {
      /* "addr" is mapped locally.  deliver msg to application */

      /* abort if invoked by refresh handler (self-refresh makes no sense) */
      /* this also to avoid deadlock on ring->write_lock */
      if (from_refresh_handler) {

#ifdef DEBUG_RW
        S(fprintf(stderr,"_neblo_UD_send(): self-refreshing!  abort...\n"));
#endif

        rc = -1;
        break;
      }

      if (pthread_mutex_trylock(&ring->write_lock)) {
        /*
         * A flush operation due to a concurrent neblo_join() is in progress.
         * When finished, the local addr interval may have changed.
         * Therefore the route for this request must be recomputed.
         */
        /* wait for the flush to end... */
        pthread_mutex_lock(&ring->write_lock);
        /* go recompute the route */
        pthread_mutex_unlock(&ring->write_lock);
        continue;
      }
      invoke_receiver_handler_UD(port,real_addr,msg,len);
      pthread_mutex_unlock(&ring->write_lock);
      break;
    }

    if (hop == (peer_t *)(-1)) {
      /*
       * "addr" is not mapped here, and this peer doesn't know where to
       * route this msg copy at the moment.
       */

      /*
       * insert the requested operation in a queue, to be tried again later
       */
      _do_(_calloc_(req,sizeof(reqst_t)),!req,"_neblo_UD_send(): req calloc");
      _do_(_calloc_(req->buf,KEYBUFLEN+SLACK),!req->buf,
           "_neblo_UD_send(): req->buf calloc");
      /* header of the write msg. see forward_msg() routine */
      sprintf(req->buf,"write %04x %08x %016llx %016llx",
              port,(u_int32_t)len,addr,real_addr);
      /* replace trailing \0 by " " */
      memset(req->buf+HEADERSIZE_W-1,32,1);
      /* put the payload after the header */
      memcpy(req->buf+HEADERSIZE_W,msg,len);
      req->th_id = 666; /* dummy th_id */
      insert_retry_queue(WRITE,req,666 /* dummy th_id */);
      rc = -1;
      break;
    }

    /* "addr" is mapped remotely; route msg through peer "hop" */

    rc = invoke_forwarder_handler_UD(port,real_addr,msg,len);
    if (rc < 0)
      /*  application decided not to forward this msg  */
      break;

    /* encapsulate "write" msg and send */
    memcpy(buf+HEADERSIZE_W,msg,len);
    rc = forward_msg(WRITE,hop,port,addr,real_addr,buf,len,0,1,
                     666 /* dummy th_id */);
    if (rc >= 0)
      break;

    /*
     * something wrong while talking to next peer, might have been flagged as
     * "tmpunavail"
     */
    tmpunavail = 1;
    /* choose another peer */

  }

  /*
   * the forward_msg() routine above might have flagged some peers as
   * "tmpunavail"...
   */
  if (tmpunavail)
    forall_peers(f,f->tmpunavail = 0);

  _free_(buf);

  if (rc < 0)
    /*  Failed to manage this msg copy.  */
    rc = -1;
  else
    rc = 0;

  if (!(cp - depth))
    /*  No more msg copies to send.  */
    return rc;

  /*  Send recursively other msg copies  */
  addr += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-depth);
  return rc + _neblo_UD_send(cp,port,addr,real_addr,msg,len,depth+1,
                             from_refresh_handler);

}



int
neblo_UD_send
(
  unsigned int cp, neblo_port port, neblo_addr addr, void *msg,
  unsigned int len
)
{
  return _neblo_UD_send(cp,port,addr,addr,msg,len,1,0);
}



/*
 *  Refresh handlers should use this variant of neblo_UD_send() to
 *  allow stored data to be re-published in the ring.
 *  Invocation outside refresh handlers is not allowed.
 */
int
neblo_UD_send_h
( neblo_port port, neblo_addr addr, void *msg, unsigned int len )
{
  if (unlikely(!__refresh_in_progress)) {
    fprintf(stderr,"neblo_UD_send_h():  not in refresh handler!\n");
    return -1;
  }
  return _neblo_UD_send(1,port,__refresh_addr,addr,msg,len,1,1);
}



/*
 * Send "cp" copies of a REQuest msg resp. to the peers who map addresses
 * "addr", "addr+2^(N-1)", "addr+2^(N-1)+2^(N-2)", ...,
 * "addr+2^(N-1)+2^(N-2)+...+2^(N-cp+1)"
 * where N is the addr size in bit (64 bit in our case).
 *
 * The msg copies are sent by successive recursive calls.  First invocation
 * must have "depth" = 1.  Each recursive call increases "depth" by 1.
 *
 * Returns -k where k is the no. of failed copies (k <= cp), so 0 on success.
 * "Failed" means it has been impossible to route msg outside of local peer;
 * they will however retried later automatically.
 *
 * Msg size cannot exceed NEBLO_MSGSIZE.
 * Semantics is "non-blocking" (split-phase), so take note of the pending
 * read messages to check for completion later.
 *
 * NOTE:  this is a support routine.  Application cannot invoke this
 * directly; they should rather use neblo_REQ_send().
 */
static int
_neblo_REQ_send
(
  int cp, neblo_port port, neblo_addr addr,
  neblo_addr real_addr, void *msg, unsigned int maxlen,
  neblo_rhandle *h, unsigned int depth
)
{

  unsigned char        *buf;
  peer_t               *hop;
  unsigned int          actual_len;
  int                   rc = 0;
  reqst_t              *req;
  unsigned int          tmpunavail = 0;


  _do_(_calloc_(buf,HEADERSIZE_R+maxlen+SLACK),!buf,
       "neblo_REQ_send(): buf calloc");
  maxlen = min(maxlen,NEBLO_MSGSIZE);

  if (depth <= 1)
    *h = (neblo_rhandle)0;

  for (;;) {

    hop = next_hop(addr);

    if (hop == NULL) {
      /* "addr" is mapped locally. deliver msg to application... */
      actual_len = maxlen;
      rc = invoke_receiver_handler_REQ(port,real_addr,msg,&actual_len);
      /*
       * this is a read msg, so the application handler may have changed
       * the content of region "msg" as well as the value of "actual_len"
       */
      if (rc >= 0) {
        /*  Application already returned the response so the read is done  */
        cancel_read_reqst(search_read_reqst(*h));
        *h = R_COMPLETE;
        break;
      }
      /*  No response, read is done however.  */
      rc = 0;
      break;
    }

    if (hop == (peer_t *)(-1)) {
      /*
       * "addr" is not mapped here, and this peer doesn't know where to
       * route this msg copy at the moment.
       */

      /*
       *  Take note of this pending read msg being forwarded, get a rhandle
       *  into "h".
       *  In case the same read msg sent in multiple copies, the
       *  routine below will reuse the same rhandle (passed through recursive
       *  calls to this routine).  No problem with this: all copies share the
       *  same real_addr.
       */
      rc = insert_pend_read(port,real_addr,msg,maxlen,NULL,R_COMPLETE,h);
      if (rc < 0)
        /*  queue of pending read msgs full.  quit.  */
        break;

      /*
       * insert the requested operation in a queue, to be tried again later
       */
      _do_(_calloc_(req,sizeof(reqst_t)),!req,"_neblo_REQ_send(): req calloc");
      _do_(_calloc_(req->buf,KEYBUFLEN+SLACK),!req->buf,
           "_neblo_REQ_send(): req->buf calloc");
      /* header of the read msg. see forward_msg() routine */
      sprintf(req->buf,"read %04x %08x %08x %016llx %016llx",
              port,(u_int32_t)maxlen,(u_int32_t)*h,addr,real_addr);
      /* replace trailing \0 by " " */
      memset(req->buf+HEADERSIZE_R-1,32,1);
      /* put payload after header.  Also read msgs can carry a payload! */
      memcpy(req->buf+HEADERSIZE_R,msg,maxlen);
      req->th_id = 666; /* dummy th_id */
      insert_retry_queue(READ,req,666 /* dummy th_id */);
      rc = -1;
      break;

    }

    /* block "addr" is mapped remotely; route msg through peer "hop" */

    rc = invoke_forwarder_handler_REQ(port,real_addr,msg,maxlen);
    if (rc < 0)
      /*  application decided not to forward this msg  */
      break;

    /*
     * take note of the pending read msg: get a rhandle into "h".
     * in case of the same read msg sent in multiple copies, the
     * routine below will reuse the same rhandle (passed through recursive
     * calls to this routine).  No problem with this: all copies share the
     * same real_addr.
     */
    rc = insert_pend_read(port,real_addr,msg,maxlen,NULL,R_COMPLETE,h);
    if (rc < 0)
      /* queue of pending read msgs full. quit */
      break;

    /* encapsulate "read" msg and send.  header carries the rhandle. */
    memcpy(buf+HEADERSIZE_R,msg,maxlen);
    rc = forward_msg(READ,hop,port,addr,real_addr,buf,maxlen,*h,1,
                     666 /* dummy th_id */);
    if (rc >= 0)
      break;

    /*
     * something wrong while talking to next peer, might have been flagged as
     * "tmpunavail"
     */
    tmpunavail = 1;
    /* choose another peer */

  }

  /*
   * the forward_msg() routine above might have flagged some peers as
   * "tmpunavail"...
   */
  if (tmpunavail)
    forall_peers(f,f->tmpunavail = 0);

  _free_(buf);

  if (rc < 0)
    /*  Failed to manage this msg copy.  */
    rc = -1;
  else
    rc = 0;


  if (*h == R_COMPLETE)
    /*  read is complete so no need to send other copies of the read msg.  */
    return rc;

  if (!(cp - depth))
    /*  No more msg copies to send.  */
    return rc;

  /*  Send recursively other msg copies  */
  addr += ((neblo_addr)1)<<(8*sizeof(neblo_addr)-depth);
  return rc + _neblo_REQ_send(cp,port,addr,real_addr,msg,maxlen,h,depth+1);

}



int
neblo_REQ_send
(
  unsigned int cp, neblo_port port, neblo_addr addr, void *msg,
  unsigned int maxlen, neblo_rhandle *h
)
{
  return _neblo_REQ_send(cp,port,addr,addr,msg,maxlen,h,1);
}



/*
 * wait for RESponse to a previously issued REQuest msg.
 * returns 0 if response is back, <0 if failure
 */
int
neblo_RES_wait
( neblo_rhandle h )
{

  pend_read_t *pr;


  if (h == R_COMPLETE)
    return 0;

  if (h == R_FAILED)
    return -1;

  pr = search_read_reqst(h);
  if (pr == NULL)
    /* pending read missing, probably timed out */
    return -1;

  pthread_mutex_lock(&pr->lock);

  for (;;) {

    if (pr->done > 0) {
      cancel_read_reqst(pr);
      pthread_mutex_unlock(&pr->lock);
      return 0;
    }

    if (!pr->in_use) {
      /* pending read vanished meanwhile, probably timed out */
      pthread_mutex_unlock(&pr->lock);
      return -1;
    }

    /* sleep on condition associated to pending read msg */
    pthread_mutex_lock(&pr->wait_sem);
    pthread_mutex_unlock(&pr->lock);
    pthread_cond_wait(&pr->wait_cond,&pr->wait_sem);
    pthread_mutex_lock(&pr->lock);
    pthread_mutex_unlock(&pr->wait_sem);
    /* woke up.  either complete, or expired */

  }

}



/*
 * test for RESponse to a previously issued REQ.
 * returns 1 if response is back, 0 if not yet, <0 if failure
 */
int
neblo_RES_test
( neblo_rhandle h )
{

  pend_read_t *pr;


  if (h == R_COMPLETE)
    return 1;

  if (h == R_FAILED)
    return -1;

  pr = (pend_read_t *)h;

  pthread_mutex_lock(&pr->lock);

  if (pr->done) {
    cancel_read_reqst(pr);
    pthread_mutex_unlock(&pr->lock);
    return 1;
  }
  if (!pr->in_use) {
    /* pending read missing, probably timed out */
    pthread_mutex_unlock(&pr->lock);
    return -1;
  }

  pthread_mutex_unlock(&pr->lock);

  return 0;

}



void
neblo_set_receiver_handler_UD
( neblo_port port, handler h )
{
  __receiver_handler_UD[port] = h;
  return;
}



void
neblo_set_receiver_handler_REQ
( neblo_port port, handler h )
{
  __receiver_handler_REQ[port] = h;
  return;
}



void
neblo_set_forwarder_handler_UD
( neblo_port port, handler h )
{
  __forwarder_handler_UD[port] = h;
  return;
}



void
neblo_set_forwarder_handler_REQ
( neblo_port port, handler h )
{
  __forwarder_handler_REQ[port] = h;
  return;
}



void
neblo_set_forwarder_handler_RES
( neblo_port port, handler h )
{
  __forwarder_handler_RES[port] = h;
  return;
}



void
neblo_set_refresh_handler
( manager h )
{
  __refresh_handler = h;
  return;
}



void
neblo_set_clean_handler
( manager h )
{
  __clean_handler = h;
  return;
}



unsigned int
neblo_in_range
( neblo_addr x, neblo_addr a, neblo_addr b )
{
  return _belong_(x,a,b);
}



void *
neblo_malloc
( unsigned int n )
{

  void *res;


  _do_(_malloc_(res,n),!res,"neblo_malloc(): failed");
  return res;

}



int
neblo_free
( void *p )
{
  _free_(p);
  return 0;
}
